@@ -11,8 +11,6 @@ import logging
 import re
 import socket
 import ssl
-from aiosmtpd import __version__
-from aiosmtpd.proxy_protocol import get_proxy, ProxyData
 from base64 import b64decode, b64encode
 from email._header_value_parser import get_addr_spec, get_angle_addr
 from email.errors import HeaderParseError
@@ -26,7 +24,9 @@ from typing import (
     List,
     NamedTuple,
     Optional,
+    Sequence,
     Tuple,
+    TypeVar,
     Union,
 )
 from warnings import warn
@@ -34,11 +34,14 @@ from warnings import warn
 import attr
 from public import public
 
+from aiosmtpd import __version__, _get_or_new_eventloop
+from aiosmtpd.proxy_protocol import ProxyData, get_proxy
+
 
 # region #### Custom Data Types #######################################################
 
 class _Missing:
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "MISSING"
 
 
@@ -58,6 +61,9 @@ AuthenticatorType = Callable[["SMTP", "Session", "Envelope", str, Any], "AuthRes
 AuthMechanismType = Callable[["SMTP", List[str]], Awaitable[Any]]
 _TriStateType = Union[None, _Missing, bytes]
 
+RT = TypeVar("RT")  # "ReturnType"
+DecoratorType = Callable[[Callable[..., RT]], Callable[..., RT]]
+
 
 # endregion
 
@@ -121,9 +127,12 @@ class AuthResult:
     """
 
     message: Optional[str] = attr.ib(kw_only=True, default=None)
-    """Optional message for additional handling by smtp_AUTH"""
+    """
+    Optional message for additional handling by smtp_AUTH.
+    Applicable only if handled == False.
+    """
 
-    auth_data: Optional[Any] = attr.ib(kw_only=True, default=None)
+    auth_data: Optional[Any] = attr.ib(kw_only=True, default=None, repr=lambda x: "...")
     """
     Optional free-form authentication data. For the built-in mechanisms, it is usually
     an instance of LoginPassword. Other implementations are free to use any data
@@ -136,10 +145,16 @@ class LoginPassword(NamedTuple):
     login: bytes
     password: bytes
 
+    def __str__(self) -> str:
+        return f"LoginPassword(login='{self.login}', password=...)"
+
+    def __repr__(self) -> str:
+        return str(self)
+
 
 @public
 class Session:
-    def __init__(self, loop):
+    def __init__(self, loop: asyncio.AbstractEventLoop):
         self.peer = None
         self.ssl = None
         self.host_name = None
@@ -162,7 +177,7 @@ class Session:
         self.authenticated = None
 
     @property
-    def login_data(self):
+    def login_data(self) -> Any:
         """Legacy login_data, usually containing the username"""
         log.warning(
             "Session.login_data is deprecated and will be removed in version 2.0"
@@ -170,7 +185,7 @@ class Session:
         return self._login_data
 
     @login_data.setter
-    def login_data(self, value):
+    def login_data(self, value: Any) -> None:
         log.warning(
             "Session.login_data is deprecated and will be removed in version 2.0"
         )
@@ -179,7 +194,7 @@ class Session:
 
 @public
 class Envelope:
-    def __init__(self):
+    def __init__(self) -> None:
         self.mail_from = None
         self.mail_options = []
         self.smtp_utf8 = False
@@ -192,12 +207,14 @@ class Envelope:
 # This is here to enable debugging output when the -E option is given to the
 # unit test suite.  In that case, this function is mocked to set the debug
 # level on the loop (as if PYTHONASYNCIODEBUG=1 were set).
-def make_loop():
-    return asyncio.get_event_loop()
+def make_loop() -> asyncio.AbstractEventLoop:
+    return _get_or_new_eventloop()
 
 
 @public
-def syntax(text, extended=None, when: Optional[str] = None):
+def syntax(
+        text: str, extended: Optional[str] = None, when: Optional[str] = None
+) -> DecoratorType:
     """
     A @decorator that provides helptext for (E)SMTP HELP.
     Applies for smtp_* methods only!
@@ -207,7 +224,7 @@ def syntax(text, extended=None, when: Optional[str] = None):
     :param when: The name of the attribute of SMTP class to check; if the value
         of the attribute is false-y then HELP will not be available for the command
     """
-    def decorator(f):
+    def decorator(f: Callable[..., RT]) -> Callable[..., RT]:
         f.__smtp_syntax__ = text
         f.__smtp_syntax_extended__ = extended
         f.__smtp_syntax_when__ = when
@@ -216,7 +233,7 @@ def syntax(text, extended=None, when: Optional[str] = None):
 
 
 @public
-def auth_mechanism(actual_name: str):
+def auth_mechanism(actual_name: str) -> DecoratorType:
     """
     A @decorator to explicitly specifies the name of the AUTH mechanism implemented by
     the function/method this decorates
@@ -224,9 +241,10 @@ def auth_mechanism(actual_name: str):
     :param actual_name: Name of AUTH mechanism. Must consists of [A-Z0-9_-] only.
         Will be converted to uppercase
     """
-    def decorator(f):
+    def decorator(f: Callable[..., RT]) -> Callable[..., RT]:
         f.__auth_mechanism_name__ = actual_name
         return f
+
     actual_name = actual_name.upper()
     if not VALID_AUTHMECH.match(actual_name):
         raise ValueError(f"Invalid AUTH mechanism name: {actual_name}")
@@ -239,7 +257,7 @@ def login_always_fail(
     return False
 
 
-def is_int(o):
+def is_int(o: Any) -> bool:
     return isinstance(o, int)
 
 
@@ -257,7 +275,7 @@ def sanitize(text: bytes) -> bytes:
 
 
 @public
-def sanitized_log(func: Callable, msg: AnyStr, *args, **kwargs):
+def sanitized_log(func: Callable[..., None], msg: AnyStr, *args, **kwargs) -> None:
     """
     Sanitize args before passing to a logging function.
     """
@@ -295,30 +313,30 @@ class SMTP(asyncio.StreamReaderProtocol):
 
     def __init__(
             self,
-            handler,
+            handler: Any,
             *,
-            data_size_limit=DATA_SIZE_DEFAULT,
-            enable_SMTPUTF8=False,
-            decode_data=False,
-            hostname=None,
-            ident=None,
+            data_size_limit: int = DATA_SIZE_DEFAULT,
+            enable_SMTPUTF8: bool = False,
+            decode_data: bool = False,
+            hostname: Optional[str] = None,
+            ident: Optional[str] = None,
             tls_context: Optional[ssl.SSLContext] = None,
-            require_starttls=False,
-            timeout=300,
-            auth_required=False,
-            auth_require_tls=True,
+            require_starttls: bool = False,
+            timeout: float = 300,
+            auth_required: bool = False,
+            auth_require_tls: bool = True,
             auth_exclude_mechanism: Optional[Iterable[str]] = None,
-            auth_callback: AuthCallbackType = None,
+            auth_callback: Optional[AuthCallbackType] = None,
             command_call_limit: Union[int, Dict[str, int], None] = None,
-            authenticator: AuthenticatorType = None,
+            authenticator: Optional[AuthenticatorType] = None,
             proxy_protocol_timeout: Optional[Union[int, float]] = None,
-            loop=None
+            loop: Optional[asyncio.AbstractEventLoop] = None
     ):
         self.__ident__ = ident or __ident__
         self.loop = loop if loop else make_loop()
         super().__init__(
             asyncio.StreamReader(loop=self.loop, limit=self.line_length_limit),
-            client_connected_cb=self._client_connected_cb,
+            client_connected_cb=self._cb_client_connected,
             loop=self.loop)
         self.event_handler = handler
         assert data_size_limit is None or isinstance(data_size_limit, int)
@@ -333,7 +351,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         self.tls_context = tls_context
         if tls_context:
             if (tls_context.verify_mode
-                    not in {ssl.CERT_NONE, ssl.CERT_OPTIONAL}):
+                    not in {ssl.CERT_NONE, ssl.CERT_OPTIONAL}):  # noqa: DUO122
                 log.warning("tls_context.verify_mode not in {CERT_NONE, "
                             "CERT_OPTIONAL}; this might cause client "
                             "connection problems")
@@ -442,13 +460,13 @@ class SMTP(asyncio.StreamReaderProtocol):
             else:
                 raise TypeError("command_call_limit must be int or Dict[str, int]")
 
-    def _create_session(self):
+    def _create_session(self) -> Session:
         return Session(self.loop)
 
-    def _create_envelope(self):
+    def _create_envelope(self) -> Envelope:
         return Envelope()
 
-    async def _call_handler_hook(self, command, *args):
+    async def _call_handler_hook(self, command: str, *args) -> Any:
         hook = self._handle_hooks.get(command)
         if hook is None:
             return MISSING
@@ -456,7 +474,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         return status
 
     @property
-    def max_command_size_limit(self):
+    def max_command_size_limit(self) -> int:
         try:
             return max(self.command_size_limits.values())
         except ValueError:
@@ -474,7 +492,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         if closed.done() and not closed.cancelled():
             closed.exception()
 
-    def connection_made(self, transport):
+    def connection_made(self, transport: asyncio.transports.BaseTransport) -> None:
         # Reset state due to rfc3207 part 4.2.
         self._set_rset_state()
         self.session = self._create_session()
@@ -503,7 +521,7 @@ class SMTP(asyncio.StreamReaderProtocol):
             self._handler_coroutine = self.loop.create_task(
                 self._handle_client())
 
-    def connection_lost(self, error):
+    def connection_lost(self, error: Optional[Exception]) -> None:
         log.info('%r connection lost', self.session.peer)
         self._timeout_handle.cancel()
         # If STARTTLS was issued, then our transport is the SSL protocol
@@ -517,7 +535,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         self._handler_coroutine.cancel()
         self.transport = None
 
-    def eof_received(self):
+    def eof_received(self) -> bool:
         log.info('%r EOF received', self.session.peer)
         self._handler_coroutine.cancel()
         if self.session.ssl is not None:
@@ -527,7 +545,7 @@ class SMTP(asyncio.StreamReaderProtocol):
             return False
         return super().eof_received()
 
-    def _reset_timeout(self, duration=None):
+    def _reset_timeout(self, duration: Optional[float] = None) -> None:
         if self._timeout_handle is not None:
             self._timeout_handle.cancel()
         self._timeout_handle = self.loop.call_later(
@@ -542,7 +560,9 @@ class SMTP(asyncio.StreamReaderProtocol):
         # up state.
         self.transport.close()
 
-    def _client_connected_cb(self, reader, writer):
+    def _cb_client_connected(
+            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
+    ):
         # This is redundant since we subclass StreamReaderProtocol, but I like
         # the shorter names.
         self._reader = reader
@@ -567,7 +587,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         log.debug("%r << %r", self.session.peer, response)
         await self._writer.drain()
 
-    async def handle_exception(self, error):
+    async def handle_exception(self, error: Exception) -> str:
         if hasattr(self.event_handler, 'handle_exception'):
             status = await self.event_handler.handle_exception(error)
             return status
@@ -668,9 +688,11 @@ class SMTP(asyncio.StreamReaderProtocol):
                             await self.push('500 Error: strict ASCII mode')
                             # Should we await self.handle_exception()?
                             continue
-                max_sz = (self.command_size_limits[command]
-                          if self.session.extended_smtp
-                          else self.command_size_limit)
+                max_sz = (
+                    self.command_size_limits[command]
+                    if self.session.extended_smtp
+                    else self.command_size_limit
+                )
                 if len(line) > max_sz:
                     await self.push('500 Command line too long')
                     continue
@@ -710,7 +732,8 @@ class SMTP(asyncio.StreamReaderProtocol):
                         self.transport.close()
                         continue
                     await self.push(
-                        '500 Error: command "%s" not recognized' % command)
+                        f'500 Error: command "{command}" not recognized'
+                    )
                     continue
 
                 # Received a valid command, reset the timer.
@@ -743,10 +766,16 @@ class SMTP(asyncio.StreamReaderProtocol):
                         status = '500 Error: Cannot describe error'
                 finally:
                     if isinstance(error, TLSSetupException):
-                        self.transport.close()
+                        # This code branch is inside None check for self.transport
+                        # so there shouldn't be a None self.transport but pytype
+                        # still complains, so silence that error.
+                        self.transport.close()  # pytype: disable=attribute-error
                         self.connection_lost(error)
                     else:
-                        await self.push(status)
+                        # The value of status is being set with ex-except and it
+                        # shouldn't be None, but pytype isn't able to infer that
+                        # so ignore the error related to wrong argument types.
+                        await self.push(status)  # pytype: disable=wrong-arg-types
 
     async def check_helo_needed(self, helo: str = "HELO") -> bool:
         """
@@ -775,7 +804,7 @@ class SMTP(asyncio.StreamReaderProtocol):
 
     # SMTP and ESMTP commands
     @syntax('HELO hostname')
-    async def smtp_HELO(self, hostname):
+    async def smtp_HELO(self, hostname: str):
         if not hostname:
             await self.push('501 Syntax: HELO hostname')
             return
@@ -788,17 +817,16 @@ class SMTP(asyncio.StreamReaderProtocol):
         await self.push(status)
 
     @syntax('EHLO hostname')
-    async def smtp_EHLO(self, hostname):
+    async def smtp_EHLO(self, hostname: str):
         if not hostname:
             await self.push('501 Syntax: EHLO hostname')
             return
 
-        response = []
+        response = ['250-' + self.hostname, ]
         self._set_rset_state()
         self.session.extended_smtp = True
-        response.append('250-%s' % self.hostname)
         if self.data_size_limit:
-            response.append('250-SIZE %s' % self.data_size_limit)
+            response.append(f'250-SIZE {self.data_size_limit}')
             self.command_size_limits['MAIL'] += 26
         if not self._decode_data:
             response.append('250-8BITMIME')
@@ -838,12 +866,12 @@ class SMTP(asyncio.StreamReaderProtocol):
             await self.push(r)
 
     @syntax('NOOP [ignored]')
-    async def smtp_NOOP(self, arg):
+    async def smtp_NOOP(self, arg: str):
         status = await self._call_handler_hook('NOOP', arg)
         await self.push('250 OK' if status is MISSING else status)
 
     @syntax('QUIT')
-    async def smtp_QUIT(self, arg):
+    async def smtp_QUIT(self, arg: str):
         if arg:
             await self.push('501 Syntax: QUIT')
         else:
@@ -853,7 +881,7 @@ class SMTP(asyncio.StreamReaderProtocol):
             self.transport.close()
 
     @syntax('STARTTLS', when='tls_context')
-    async def smtp_STARTTLS(self, arg):
+    async def smtp_STARTTLS(self, arg: str):
         if arg:
             await self.push('501 Syntax: STARTTLS')
             return
@@ -871,11 +899,12 @@ class SMTP(asyncio.StreamReaderProtocol):
             self.tls_context,
             waiter,
             server_side=True)
+
         # Reconfigure transport layer.  Keep a reference to the original
         # transport so that we can close it explicitly when the connection is
-        # lost.  XXX BaseTransport.set_protocol() was added in Python 3.5.3 :(
+        # lost.
         self._original_transport = self.transport
-        self._original_transport._protocol = self._tls_protocol
+        self._original_transport.set_protocol(self._tls_protocol)
         # Reconfigure the protocol layer.  Why is the app transport a protected
         # property, if it MUST be used externally?
         self.transport = self._tls_protocol._app_transport
@@ -915,14 +944,15 @@ class SMTP(asyncio.StreamReaderProtocol):
         status = await self._call_handler_hook('AUTH', args)
         if status is MISSING:
             auth_method = self._auth_methods[mechanism]
-            if auth_method.is_builtin:
-                log.debug(f"Using builtin auth_ hook for {mechanism}")
-            else:
-                log.debug(f"Using handler auth_ hook for {mechanism}")
+            log.debug(
+                "Using %s auth_ hook for %r",
+                "builtin" if auth_method.is_builtin else "handler",
+                mechanism
+            )
             # Pass 'self' to method so external methods can leverage this
             # class's helper methods such as push()
             auth_result = await auth_method.method(self, args)
-            log.debug(f"auth_{mechanism} returned {auth_result}")
+            log.debug("auth_%s returned %r", mechanism, auth_result)
 
             # New system using `authenticator` and AuthResult
             if isinstance(auth_result, AuthResult):
@@ -952,7 +982,6 @@ class SMTP(asyncio.StreamReaderProtocol):
                 # is rejected / not valid
                 status = CODE_INVALID
             else:
-
                 self.session.login_data = auth_result
                 status = CODE_SUCCESS
 
@@ -987,7 +1016,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         challenge = b"334 " + (b64encode(challenge) if encode_to_b64 else challenge)
         log.debug("%r << challenge: %r", self.session.peer, challenge)
         await self.push(challenge)
-        line = await self._reader.readline()
+        line = await self._reader.readline()      # pytype: disable=attribute-error
         if log_client_response:
             warn("AUTH interaction logging is enabled!")
             warn("Sensitive information might be leaked!")
@@ -1022,7 +1051,7 @@ class SMTP(asyncio.StreamReaderProtocol):
             encode_to_b64=False,
         )
 
-    def _authenticate(self, mechanism, auth_data) -> AuthResult:
+    def _authenticate(self, mechanism: str, auth_data: Any) -> AuthResult:
         if self._authenticator is not None:
             # self.envelope is likely still empty, but we'll pass it anyways to
             # make the invocation similar to the one in _call_handler_hook
@@ -1074,7 +1103,7 @@ class SMTP(asyncio.StreamReaderProtocol):
             # login data is "{authz_id}\x00{login_id}\x00{password}"
             # authz_id can be null, and currently ignored
             # See https://tools.ietf.org/html/rfc4616#page-3
-            _, login, password = login_and_password.split(b"\x00")
+            _, login, password = login_and_password.split(b"\x00")  # pytype: disable=attribute-error  # noqa: E501
         except ValueError:  # not enough args
             await self.push("501 5.5.2 Can't split auth value")
             return AuthResult(success=False, handled=True)
@@ -1083,7 +1112,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         assert password is not None
         return self._authenticate("PLAIN", LoginPassword(login, password))
 
-    async def auth_LOGIN(self, _, args: List[str]):
+    async def auth_LOGIN(self, _, args: List[str]) -> AuthResult:
         login: _TriStateType
         if len(args) == 1:
             # Client sent only "AUTH LOGIN"
@@ -1107,13 +1136,13 @@ class SMTP(asyncio.StreamReaderProtocol):
 
         return self._authenticate("LOGIN", LoginPassword(login, password))
 
-    def _strip_command_keyword(self, keyword, arg):
+    def _strip_command_keyword(self, keyword: str, arg: str) -> Optional[str]:
         keylen = len(keyword)
         if arg[:keylen].upper() == keyword:
             return arg[keylen:].strip()
         return None
 
-    def _getaddr(self, arg) -> Tuple[Optional[str], Optional[str]]:
+    def _getaddr(self, arg: str) -> Tuple[Optional[str], Optional[str]]:
         """
         Try to parse address given in SMTP command.
 
@@ -1135,7 +1164,9 @@ class SMTP(asyncio.StreamReaderProtocol):
             return None, None
         return address, rest
 
-    def _getparams(self, params):
+    def _getparams(
+            self, params: Sequence[str]
+    ) -> Optional[Dict[str, Union[str, bool]]]:
         # Return params as dictionary. Return None if not all parameters
         # appear to be syntactically valid according to RFC 1869.
         result = {}
@@ -1146,7 +1177,8 @@ class SMTP(asyncio.StreamReaderProtocol):
             result[param] = value if eq else True
         return result
 
-    def _syntax_available(self, method):
+    # noinspection PyUnresolvedReferences
+    def _syntax_available(self, method: Callable) -> bool:
         if not hasattr(method, '__smtp_syntax__'):
             return False
         if method.__smtp_syntax_when__:
@@ -1183,7 +1215,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         if arg:
             address, params = self._getaddr(arg)
             if address is None:
-                await self.push('502 Could not VRFY %s' % arg)
+                await self.push('502 Could not VRFY ' + arg)
             else:
                 status = await self._call_handler_hook('VRFY', address)
                 await self.push(
@@ -1304,7 +1336,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         await self.push(status)
 
     @syntax('RSET')
-    async def smtp_RSET(self, arg):
+    async def smtp_RSET(self, arg: str):
         if arg:
             await self.push('501 Syntax: RSET')
             return
@@ -1448,5 +1480,5 @@ class SMTP(asyncio.StreamReaderProtocol):
         await self.push('250 OK' if status is MISSING else status)
 
     # Commands that have not been implemented.
-    async def smtp_EXPN(self, arg):
+    async def smtp_EXPN(self, arg: str):
         await self.push('502 EXPN not implemented')
